# -*- coding: utf-8 -*-
# Generated by Django 1.11.11 on 2018-03-19 14:46
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('message_sender', '0018_auto_20180316_1305'),
    ]

    operations = [
        migrations.CreateModel(
            name='ArchivedOutbounds',
            fields=[
                ('date', models.DateField(help_text=b'The date that the archive is for', primary_key=True, serialize=False)),
                ('archive', models.FileField(help_text=b'The file for the archive', upload_to=b'')),
            ],
            options={
                'verbose_name': 'archived outbounds',
                'verbose_name_plural': 'archived outbounds',
            },
        ),
        migrations.AlterModelOptions(
            name='aggregateoutbounds',
            options={'verbose_name': 'aggregate outbounds', 'verbose_name_plural': 'aggregate outbounds'},
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.12 on 2017-05-03 12:48
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('message_sender', '0013_outbound_channel'),
    ]

    operations = [
        migrations.AddField(
            model_name='outbound',
            name='to_identity',
            field=models.CharField(blank=True, max_length=36, null=False, db_index=True),
        ),
        migrations.AlterField(
            model_name='outbound',
            name='to_addr',
            field=models.CharField(blank=True, db_index=True, max_length=500, null=False),
        ),
        migrations.AddField(
            model_name='inbound',
            name='from_identity',
            field=models.CharField(blank=True, max_length=36, null=False, db_index=True, default=''),
        ),
        migrations.AlterField(
            model_name='inbound',
            name='from_addr',
            field=models.CharField(blank=True, db_index=True, max_length=255, null=False),
        ),
    ]



from django.contrib import admin
from django import forms
from django.core.exceptions import ValidationError
from django.core.paginator import Paginator
from django.db import connections
from django.utils.functional import cached_property

from .models import (
    Outbound, Inbound, Channel, AggregateOutbounds, ArchivedOutbounds,
)
from .tasks import send_message


class EstimatedCountPaginator(Paginator):
    def __init__(self, *args, **kwargs):
        super(EstimatedCountPaginator, self).__init__(*args, **kwargs)
        self.object_list.count = self.count

    @cached_property
    def count(self):
        if self.object_list.query.where:
            return self.object_list.count()

        db_table = self.object_list.model._meta.db_table
        cursor = connections[self.object_list.db].cursor()
        cursor.execute(
            "SELECT reltuples FROM pg_class WHERE relname = %s", (db_table, ))
        result = cursor.fetchone()
        if not result:
            return 0
        return int(result[0])


class OutboundAdmin(admin.ModelAdmin):
    list_display = ('to_addr', 'to_identity', 'delivered', 'attempts',
                    'vumi_message_id', 'created_at', 'updated_at', 'content', )
    list_display_links = ('to_addr', 'to_identity')
    list_filter = ('delivered', 'attempts', 'created_at', 'updated_at', )
    search_fields = ['to_addr', 'to_identity']
    actions = ["resend_outbound"]
    paginator = EstimatedCountPaginator

    def resend_outbound(self, request, queryset):
        resent = 0
        for record in queryset.iterator():
            send_message.apply_async(kwargs={"message_id": str(record.id)})
            resent += 1
        if resent == 1:
            output_text = "message"
        else:
            output_text = "messages"
        self.message_user(request, "Attempting to resend %s %s." % (
                          resent, output_text))

    resend_outbound.short_description = (
        "Resend selected outbounds")


class InboundAdmin(admin.ModelAdmin):
    list_display = ('message_id', 'in_reply_to', 'to_addr', 'from_addr',
                    'from_identity', 'created_at', 'updated_at', 'content', )
    list_filter = ('in_reply_to', 'from_addr', 'created_at', 'updated_at', )
    search_fields = ['to_addr', 'from_identity']
    paginator = EstimatedCountPaginator


class ChannelAdminForm(forms.ModelForm):
    def clean(self):
        channel_id = self.cleaned_data['channel_id']
        channel_type = self.cleaned_data['channel_type']
        config = self.cleaned_data['configuration']

        missing = []
        if channel_type == Channel.JUNEBUG_TYPE:
            keys = ('JUNEBUG_API_URL', 'JUNEBUG_API_AUTH', 'JUNEBUG_API_FROM')

        elif channel_type == Channel.VUMI_TYPE:
            keys = ('VUMI_CONVERSATION_KEY', 'VUMI_ACCOUNT_KEY',
                    'VUMI_ACCOUNT_TOKEN', 'VUMI_API_URL')
        elif channel_type == Channel.HTTP_API_TYPE:
            keys = ('HTTP_API_URL', 'HTTP_API_AUTH', 'HTTP_API_FROM')

        for key in keys:
            if key not in config.keys():
                missing.append(key)

        if missing:
            raise ValidationError(
                "Configuration keys missing: {}".format(', '.join(missing)))

        if not self.cleaned_data['default']:
            if not Channel.objects.filter(default=True).exclude(
                    channel_id=channel_id).exists():
                raise ValidationError(
                    "Please make sure there is a default channel.")

        return self.cleaned_data


class ChannelAdmin(admin.ModelAdmin):
    list_display = ('channel_id', 'channel_type', 'concurrency_limit',
                    'default')
    list_filter = ('channel_type', 'default')
    search_fields = ['channel_id']
    form = ChannelAdminForm


class AggregateOutboundsAdmin(admin.ModelAdmin):
    list_display = (
        'date', 'delivered', 'channel', 'attempts', 'total',
    )
    list_filter = (
        'date', 'delivered', 'channel',
    )
    paginator = EstimatedCountPaginator


class ArchivedOutboundsAdmin(admin.ModelAdmin):
    list_display = (
        'date', 'archive',
    )
    list_filter = (
        'date',
    )

admin.site.register(Outbound, OutboundAdmin)
admin.site.register(Inbound, InboundAdmin)
admin.site.register(Channel, ChannelAdmin)
admin.site.register(AggregateOutbounds, AggregateOutboundsAdmin)
admin.site.register(ArchivedOutbounds, ArchivedOutboundsAdmin)



from django.conf import settings
from importlib import import_module
from seed_services_client import IdentityStoreApiClient

identity_store_client = IdentityStoreApiClient(
    settings.IDENTITY_STORE_TOKEN,
    settings.IDENTITY_STORE_URL,
    retries=5,
    timeout=settings.DEFAULT_REQUEST_TIMEOUT,
)


def get_available_metrics():
    available_metrics = []
    available_metrics.extend(settings.METRICS_REALTIME)
    available_metrics.extend(settings.METRICS_SCHEDULED)

    return available_metrics


def load_callable(dotted_path_to_callable):
    module_name, func_name = dotted_path_to_callable.rsplit('.', 1)
    mod = import_module(module_name)
    func = getattr(mod, func_name)
    return func


def get_identity_address(identity_uuid, use_communicate_through=False):
    params = {"default": True}
    if use_communicate_through:
        params['use_communicate_through'] = True

    return identity_store_client.get_identity_address(
        identity_uuid, params=params)


def get_identity_by_address(address_value, address_type="msisdn"):
    r = identity_store_client.get_identity_by_address(address_type,
                                                      address_value)

    results = list(r["results"])
    if len(results) > 0:
        return results
    else:
        return None


def create_identity(identity):
    return identity_store_client.create_identity(identity)



# -*- coding: utf-8 -*-
# Generated by Django 1.9.12 on 2017-04-19 11:56
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('message_sender', '0007_add_to_addr_index'),
        ('message_sender', '0009_auto_20170323_1453'),
    ]

    operations = [
    ]



from datetime import datetime, timedelta
from django.core.exceptions import ObjectDoesNotExist, MultipleObjectsReturned
from django.conf import settings
from django.contrib.auth.models import User
from django import forms
from rest_hooks.models import Hook
from rest_framework import viewsets, status, filters, mixins
from rest_framework.pagination import CursorPagination
from rest_framework.permissions import IsAuthenticated, AllowAny, IsAdminUser
from rest_framework.views import APIView
from rest_framework.response import Response
from rest_framework.authtoken.models import Token

from .models import (
    Outbound, Inbound, OutboundSendFailure, Channel, InvalidMessage,
    AggregateOutbounds, ArchivedOutbounds)
from .serializers import (
    OutboundSerializer, InboundSerializer, JunebugInboundSerializer,
    HookSerializer, CreateUserSerializer, OutboundSendFailureSerializer,
    AggregateOutboundSerializer, ArchivedOutboundSerializer,
)
from .tasks import (
    send_message, fire_metric, ConcurrencyLimiter, requeue_failed_tasks,
    aggregate_outbounds, archive_outbound,
)
from seed_message_sender.utils import (
    get_available_metrics, get_identity_by_address, create_identity)
from seed_papertrail.decorators import papertrail
import django_filters

# Uncomment line below if scheduled metrics are added
# from .tasks import scheduled_metrics


class IdCursorPagination(CursorPagination):
    ordering = "-id"


class UserView(APIView):
    """ API endpoint that allows users creation and returns their token.
    Only admin users can do this to avoid permissions escalation.
    """
    permission_classes = (IsAdminUser,)

    def post(self, request):
        '''Create a user and token, given an email. If user exists just
        provide the token.'''
        serializer = CreateUserSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        email = serializer.validated_data.get('email')
        try:
            user = User.objects.get(username=email)
        except User.DoesNotExist:
            user = User.objects.create_user(email, email=email)
        token, created = Token.objects.get_or_create(user=user)

        return Response(
            status=status.HTTP_201_CREATED, data={'token': token.key})


class HookViewSet(viewsets.ModelViewSet):
    """
    Retrieve, create, update or destroy webhooks.
    """
    permission_classes = (IsAuthenticated,)
    queryset = Hook.objects.all()
    serializer_class = HookSerializer

    def perform_create(self, serializer):
        serializer.save(user=self.request.user)


class MultipleField(forms.Field):
    widget = forms.MultipleHiddenInput

    def clean(self, value):
        if value is None:
            return None
        return [super(MultipleField, self).clean(v) for v in value]


class MultipleFilter(django_filters.Filter):
    field_class = MultipleField

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('lookup_expr', 'in')
        super(MultipleFilter, self).__init__(*args, **kwargs)


class OutboundFilter(filters.FilterSet):
    before = django_filters.IsoDateTimeFilter(name="created_at",
                                              lookup_expr='lte')
    after = django_filters.IsoDateTimeFilter(name="created_at",
                                             lookup_expr='gte')
    to_addr = MultipleFilter(name='to_addr')
    to_identity = MultipleFilter(name='to_identity')

    class Meta:
        model = Outbound
        fields = ('version', 'vumi_message_id',
                  'delivered', 'attempts',
                  'created_at', 'updated_at',
                  'before', 'after')


class OutboundViewSet(viewsets.ModelViewSet):
    """
    API endpoint that allows Outbound models to be viewed or edited.
    """
    permission_classes = (IsAuthenticated,)
    queryset = Outbound.objects.all()
    serializer_class = OutboundSerializer
    filter_class = OutboundFilter
    filter_backends = (filters.DjangoFilterBackend, filters.OrderingFilter)
    ordering_fields = ('created_at',)
    ordering = ('-created_at',)

    @papertrail.debug('api_outbound_create', sample=0.1)
    def create(self, *args, **kwargs):
        return super(OutboundViewSet, self).create(*args, **kwargs)


class InboundFilter(filters.FilterSet):
    from_addr = MultipleFilter(name='from_addr')
    from_identity = MultipleFilter(name='from_identity')

    class Meta:
        model = Inbound
        fields = (
            'message_id', 'in_reply_to', 'to_addr', 'content',
            'transport_name', 'transport_type', 'created_at', 'updated_at',)


class InboundViewSet(viewsets.ModelViewSet):

    """
    API endpoint that allows Inbound models to be viewed or edited.
    """
    permission_classes = (IsAuthenticated,)
    queryset = Inbound.objects.all()
    filter_class = InboundFilter
    filter_backends = (filters.DjangoFilterBackend, filters.OrderingFilter)
    ordering_fields = ('created_at',)
    ordering = ('-created_at',)

    def get_serializer_class(self):
        if self.action == 'create':
            if "channel_data" in self.request.data:
                return JunebugInboundSerializer
        return InboundSerializer

    def create(self, request, *args, **kwargs):
        if not kwargs.get('channel_id'):
            channel = Channel.objects.get(default=True)
        else:
            channel = Channel.objects.get(channel_id=kwargs.get('channel_id'))

        if "from" in request.data:
            msisdn = request.data.pop("from")
        elif "from_addr" in request.data:
            msisdn = request.data.pop("from_addr")

        result = get_identity_by_address(msisdn)

        if result:
            identity_id = result[0]['id']
        else:
            identity = {
                'details': {
                    'default_addr_type': 'msisdn',
                    'addresses': {
                        'msisdn': {
                            msisdn: {'default': True}
                        }
                    }
                }
            }
            identity = create_identity(identity)
            identity_id = identity['id']

        request.data['from_identity'] = identity_id

        if channel.concurrency_limit == 0:
            return super(InboundViewSet, self).create(request, *args, **kwargs)

        close_event = False
        # Handle message from Junebug
        if request.data.get("channel_data", {}).get("session_event", None) == \
                "close":
            close_event = True
            related_outbound = request.data["reply_to"]
        elif "session_event" in request.data:  # Handle message from Vumi
            if request.data["session_event"] == "close":
                close_event = True
                related_outbound = request.data["in_reply_to"]

        if close_event:
            if related_outbound is not None:
                try:
                    message = Outbound.objects.get(
                        vumi_message_id=related_outbound)
                except (ObjectDoesNotExist, MultipleObjectsReturned):
                    message = Outbound.objects.filter(
                        to_identity=identity_id).order_by('-created_at').last()
            else:
                message = Outbound.objects.filter(
                    to_identity=identity_id).order_by('-created_at').last()
            if message:
                ConcurrencyLimiter.decr_message_count(
                    channel, message.last_sent_time)

        return super(InboundViewSet, self).create(request, *args, **kwargs)


def fire_delivery_hook(outbound):
    outbound.refresh_from_db()
    # Only fire if the message has been delivered or we've reached max attempts
    if (not outbound.delivered and
            outbound.attempts < settings.MESSAGE_SENDER_MAX_RETRIES):
        return

    payload = {
        'outbound_id': str(outbound.id),
        'delivered': outbound.delivered,
        'to_addr': outbound.to_addr,
    }
    if hasattr(outbound, 'to_identity'):
        payload['identity'] = outbound.to_identity

    if payload['to_addr'] is None and payload.get('identity', None) is None:
        raise InvalidMessage(outbound)

    # Becaues the Junebug event endpoint has no authentication, we get an
    # AnonymousUser object for the user. So we have to manually find all of the
    # hook events, ignoring the user, and deliver them.
    hooks = Hook.objects.filter(event='outbound.delivery_report')
    for hook in hooks:
        hook.deliver_hook(None, payload_override={
            'hook': hook.dict(),
            'data': payload,
            }
        )


def decr_message_count(message):
    if message.channel:
        channel = message.channel
    else:
        channel = Channel.objects.get(default=True)
    ConcurrencyLimiter.decr_message_count(
        channel, message.last_sent_time)


class EventListener(APIView):

    """
    Triggers updates to outbound messages based on event data from Vumi
    """
    permission_classes = (AllowAny,)

    def post(self, request, *args, **kwargs):
        """
        Checks for expect event types before continuing
        """

        try:
            expect = ["message_type", "event_type", "user_message_id",
                      "event_id", "timestamp"]
            if set(expect).issubset(request.data.keys()):
                # Load message
                message = Outbound.objects.select_related('channel').get(
                    vumi_message_id=request.data["user_message_id"])
                # only expecting `event` on this endpoint
                if request.data["message_type"] == "event":
                    event = request.data["event_type"]
                    # expecting ack, nack, delivery_report
                    if event == "ack":
                        message.delivered = True
                        message.to_addr = ''
                        message.metadata["ack_timestamp"] = \
                            request.data["timestamp"]
                        message.save()
                        fire_delivery_hook(message)

                        # OBD number of successful tries metric
                        if "voice_speech_url" in message.metadata:
                            fire_metric.apply_async(kwargs={
                                "metric_name":
                                    'vumimessage.obd.successful.sum',
                                "metric_value": 1.0
                            })
                    elif event == "delivery_report":
                        message.delivered = True
                        message.to_addr = ''
                        message.metadata["delivery_timestamp"] = \
                            request.data["timestamp"]
                        message.save()
                        fire_delivery_hook(message)
                    elif event == "nack":
                        if "nack_reason" in request.data:
                            message.metadata["nack_reason"] = \
                                request.data["nack_reason"]
                            message.save()
                        fire_delivery_hook(message)

                        decr_message_count(message)

                        send_message.delay(str(message.id))
                        if "voice_speech_url" in message.metadata:
                            fire_metric.apply_async(kwargs={
                                "metric_name":
                                    'vumimessage.obd.unsuccessful.sum',
                                "metric_value": 1.0
                            })

                    # Return
                    status = 200
                    accepted = {"accepted": True}
                else:
                    status = 400
                    accepted = {"accepted": False,
                                "reason": "Unexpected message_type"}
            else:
                status = 400
                accepted = {"accepted": False,
                            "reason": "Missing expected body keys"}
        except ObjectDoesNotExist:
            status = 400
            accepted = {"accepted": False,
                        "reason": "Missing message in control"}
        return Response(accepted, status=status)

    # TODO make this work in test harness, works in production
    # def perform_create(self, serializer):
    #     serializer.save(created_by=self.request.user,
    #                     updated_by=self.request.user)
    #
    # def perform_update(self, serializer):
    #     serializer.save(updated_by=self.request.user)


class JunebugEventListener(APIView):

    """
    Triggers updates to outbound messages based on event data from Junebug
    """
    permission_classes = (AllowAny,)

    def post(self, request, *args, **kwargs):
        """
        Updates the message from the event data.
        """
        expect = ["event_type", "message_id", "timestamp"]
        if not set(expect).issubset(request.data.keys()):
            return Response({
                "accepted": False,
                "reason": "Missing expected body keys"
            }, status=400)

        try:
            message = Outbound.objects.select_related('channel').get(
                vumi_message_id=request.data["message_id"])
        except ObjectDoesNotExist:
            return Response({
                "accepted": False,
                "reason": "Cannot find message for event"
            }, status=400)

        event_type = request.data["event_type"]
        if event_type == "submitted":
            message.delivered = True
            message.to_addr = ''
            message.metadata["ack_timestamp"] = request.data["timestamp"]
            message.save(update_fields=['metadata', 'delivered', 'to_addr'])
            fire_delivery_hook(message)

            # OBD number of successful tries metric
            if "voice_speech_url" in message.metadata:
                fire_metric.apply_async(kwargs={
                    "metric_name": 'vumimessage.obd.successful.sum',
                    "metric_value": 1.0
                })
        elif event_type == "rejected":
            message.metadata["nack_reason"] = (
                request.data.get("event_details"))
            message.save(update_fields=['metadata'])
            fire_delivery_hook(message)
            decr_message_count(message)
            send_message.delay(str(message.id))
        elif event_type == "delivery_succeeded":
            message.delivered = True
            message.to_addr = ''
            message.metadata["delivery_timestamp"] = request.data["timestamp"]
            message.save(update_fields=['delivered', 'metadata', 'to_addr'])
            fire_delivery_hook(message)
        elif event_type == "delivery_failed":
            message.metadata["delivery_failed_reason"] = (
                request.data.get("event_details"))
            message.save(update_fields=['metadata'])
            fire_delivery_hook(message)
            decr_message_count(message)
            send_message.delay(str(message.id))

        if ("voice_speech_url" in message.metadata and
                event_type in ("rejected", "delivery_failed")):
            fire_metric.apply_async(kwargs={
                "metric_name": 'vumimessage.obd.unsuccessful.sum',
                "metric_value": 1.0
            })

        return Response({"accepted": True}, status=200)


class MetricsView(APIView):

    """ Metrics Interaction
        GET - returns list of all available metrics on the service
        POST - starts up the task that fires all the scheduled metrics
    """
    permission_classes = (IsAuthenticated,)

    def get(self, request, *args, **kwargs):
        status = 200
        resp = {
            "metrics_available": get_available_metrics()
        }
        return Response(resp, status=status)

    def post(self, request, *args, **kwargs):
        status = 201
        # Uncomment line below if scheduled metrics are added
        # scheduled_metrics.apply_async()
        resp = {"scheduled_metrics_initiated": True}
        return Response(resp, status=status)


class HealthcheckView(APIView):

    """ Healthcheck Interaction
        GET - returns service up - getting auth'd requires DB
    """
    permission_classes = (IsAuthenticated,)

    def get(self, request, *args, **kwargs):
        import seed_message_sender
        import django
        import rest_framework
        status = 200
        resp = {
            "up": True,
            "result": {
                "database": "Accessible",
                "version": seed_message_sender.__version__,
                "libraries": {
                    "django": django.__version__,
                    "djangorestframework": rest_framework.__version__
                }
            }
        }
        return Response(resp, status=status)


class FailedTaskViewSet(mixins.ListModelMixin,
                        mixins.RetrieveModelMixin,
                        viewsets.GenericViewSet):
    permission_classes = (IsAuthenticated,)
    queryset = OutboundSendFailure.objects.all()
    serializer_class = OutboundSendFailureSerializer
    pagination_class = IdCursorPagination

    def create(self, request):
        status = 201
        resp = {'requeued_failed_tasks': True}
        requeue_failed_tasks.delay()
        return Response(resp, status=status)


class AggregateOutboundViewSet(viewsets.GenericViewSet):
    permission_classes = (IsAuthenticated,)
    queryset = AggregateOutbounds.objects.all()
    serializer_class = AggregateOutboundSerializer

    def create(self, request):
        serializer = self.get_serializer_class()(data=request.data)
        serializer.is_valid(raise_exception=True)
        start = serializer.validated_data.get('start', None)
        end = serializer.validated_data.get('end', None)
        if not end:
            end = datetime.now().date()
        if not start:
            diff = timedelta(days=settings.AGGREGATE_OUTBOUND_BACKTRACK)
            start = (datetime.now() - diff).date()
        aggregate_outbounds.delay(
            start.isoformat(), end.isoformat())
        return Response({'aggregate_outbounds': True}, status=202)


class ArchivedOutboundViewSet(viewsets.GenericViewSet):
    permission_classes = (IsAuthenticated,)
    queryset = ArchivedOutbounds.objects.all()
    serializer_class = ArchivedOutboundSerializer

    def create(self, request):
        serializer = self.get_serializer_class()(data=request.data)
        serializer.is_valid(raise_exception=True)
        start = serializer.validated_data['start']
        end = serializer.validated_data['end']
        archive_outbound.delay(start.isoformat(), end.isoformat())
        return Response({'archived_outbounds': True}, status=202)



from datetime import timedelta
from django.conf import settings
from django.contrib.sites.shortcuts import get_current_site

try:
    from urlparse import urlunparse
except ImportError:
    from urllib.parse import urlunparse


def make_absolute_url(path):
    # NOTE: We're using the default site as set by
    #       settings.SITE_ID and the Sites framework
    site = get_current_site(None)
    return urlunparse(
        ('https' if settings.USE_SSL else 'http',
         site.domain, path,
         '', '', ''))


def daterange(start_date, end_date):
    """
    Returns the dates between the start and end dates
    """
    for n in range(int((end_date - start_date).days) + 1):
        yield start_date + timedelta(n)



import uuid

from django.contrib.postgres.fields import JSONField
from django.contrib.auth.models import User
from django.db import models
from django.utils.encoding import python_2_unicode_compatible


@python_2_unicode_compatible
class Channel(models.Model):

    VUMI_TYPE = 'vumi'
    JUNEBUG_TYPE = 'junebug'
    HTTP_API_TYPE = 'http_api'

    CHANNEL_TYPES = (
        (JUNEBUG_TYPE, 'Junebug'),
        (VUMI_TYPE, 'Vumi'),
        (HTTP_API_TYPE, 'HTTP API')
    )

    channel_id = models.CharField(primary_key=True, editable=True,
                                  max_length=64)
    channel_type = models.CharField(choices=CHANNEL_TYPES, max_length=20,
                                    default=JUNEBUG_TYPE)
    concurrency_limit = models.IntegerField(null=False, blank=False, default=0)
    message_delay = models.IntegerField(null=False, blank=False, default=0)
    message_timeout = models.IntegerField(null=False, blank=False, default=0)
    default = models.BooleanField(default=False)
    configuration = JSONField()

    def __str__(self):  # __unicode__ on Python 2
        return str(self.channel_id)


class InvalidMessage(Exception):
    """
    The message that has been stored in the database is not a valid message.
    """
    def __init__(self, message):
        return super(InvalidMessage, self).__init__(
            'Invalid message: {}'.format(message.id))


@python_2_unicode_compatible
class Outbound(models.Model):

    """
    Contacts outbound messages and their status.
    Delivered is set to true when ack received because delivery reports patchy
    """
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    to_addr = models.CharField(null=False, blank=True, max_length=500,
                               db_index=True)
    to_identity = models.CharField(max_length=36, null=False, blank=True,
                                   db_index=True)
    version = models.IntegerField(default=1)
    content = models.CharField(null=True, blank=True, max_length=1000)
    vumi_message_id = models.CharField(null=True, blank=True, max_length=36,
                                       db_index=True)
    delivered = models.BooleanField(default=False)
    resend = models.NullBooleanField(
        default=None, null=True, blank=True, help_text="True if this is a "
        "resend requested by the user.")
    call_answered = models.NullBooleanField(
        default=None, null=True, blank=True, help_text="True if the call has "
        "been answered. Not used for text messages")
    attempts = models.IntegerField(default=0)
    metadata = JSONField()
    channel = models.ForeignKey(Channel, null=True)
    updated_at = models.DateTimeField(auto_now=True)
    created_at = models.DateTimeField(auto_now_add=True, db_index=True)
    last_sent_time = models.DateTimeField(null=True, blank=True)
    created_by = models.ForeignKey(User, related_name='outbounds_created',
                                   null=True)
    updated_by = models.ForeignKey(User, related_name='outbounds_updated',
                                   null=True)
    user = property(lambda self: self.created_by)

    def __str__(self):  # __unicode__ on Python 2
        return str(self.id)


@python_2_unicode_compatible
class Inbound(models.Model):

    """
    Contacts inbound messages from Vumi
    """
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    message_id = models.CharField(null=False, blank=False, max_length=36)
    in_reply_to = models.CharField(null=True, blank=True, max_length=36)
    to_addr = models.CharField(null=False, blank=False, max_length=255)
    from_addr = models.CharField(
        null=False, blank=True, max_length=255, db_index=True)
    from_identity = models.CharField(max_length=36, null=False, blank=True,
                                     db_index=True, default="")
    content = models.CharField(null=True, blank=True, max_length=1000)
    transport_name = models.CharField(null=False, blank=False, max_length=200)
    transport_type = models.CharField(null=True, blank=True, max_length=200)
    helper_metadata = JSONField()
    updated_at = models.DateTimeField(auto_now=True)
    created_at = models.DateTimeField(auto_now_add=True, db_index=True)
    created_by = models.ForeignKey(User, related_name='inbounds_created',
                                   null=True)
    updated_by = models.ForeignKey(User, related_name='inbounds_updated',
                                   null=True)
    user = property(lambda self: self.created_by)

    def __str__(self):  # __unicode__ on Python 2
        return str(self.id)


@python_2_unicode_compatible
class OutboundSendFailure(models.Model):
    outbound = models.ForeignKey(Outbound, on_delete=models.CASCADE)
    task_id = models.UUIDField()
    initiated_at = models.DateTimeField()
    reason = models.TextField()

    def __str__(self):  # __unicode__ on Python 2
        return str(self.id)


@python_2_unicode_compatible
class IdentityLookup(models.Model):

    msisdn = models.CharField(primary_key=True, max_length=255)
    identity = models.CharField(max_length=36, null=False, blank=False)

    def __str__(self):  # __unicode__ on Python 2
        return str(self.identity)


@python_2_unicode_compatible
class AggregateOutbounds(models.Model):
    date = models.DateField(
        help_text="The date that the aggregate is for")
    delivered = models.BooleanField(
        help_text="Whether this is for delivery passed or failed messages")
    channel = models.ForeignKey(
        to=Channel, on_delete=models.SET_NULL, null=True,
        help_text="Which channel this is for")
    attempts = models.IntegerField(
        help_text="The total number of attempts")
    total = models.IntegerField(
        help_text="The total number of messages")

    class Meta:
        unique_together = [
            ['date', 'delivered', 'channel'],
        ]
        verbose_name = "aggregate outbounds"
        verbose_name_plural = "aggregate outbounds"

    def __str__(self):
        return "{}: {} {}".format(
            self.date, self.channel_id,
            "delivered" if self.delivered else "not delivered",
        )


@python_2_unicode_compatible
class ArchivedOutbounds(models.Model):
    date = models.DateField(
        help_text="The date that the archive is for", primary_key=True)
    archive = models.FileField(
        help_text="The file for the archive")

    class Meta:
        verbose_name = "archived outbounds"
        verbose_name_plural = "archived outbounds"

    def __str__(self):
        return "{}: {}".format(self.date, self.archive)



import os
import csv

from django.core.management.base import BaseCommand, CommandError
from django.apps import apps


class Command(BaseCommand):
    args = 'Model.csv'
    help = 'Import `Model`.csv into `Model` database.'

    def handle(self, *args, **options):
        if len(args) != 1:
            raise CommandError("Invalid Invocation. See help.")

        csvPath = args[0]
        if not os.path.exists(csvPath):
            raise CommandError("%s doesnt exist." % csvPath)

        model, _ = os.path.splitext(os.path.basename(csvPath))
        Model = apps.get_model("message_sender", model.title())
        if not Model:
            raise CommandError("%s Model doesn't exist'")

        model_fields = [f.name for f in Model._meta.fields]
        fields_name = []
        with open(csvPath, 'rb') as csvFile:
            reader = csv.reader(csvFile, delimiter=',', quotechar="\"")
            fields_name = reader.next()
            for i, _ in enumerate(fields_name):
                fields_name[i] = fields_name[i].lower()
                fields_name[i] = fields_name[i].replace(' ', '_')
                if not fields_name[i] in model_fields:
                    raise CommandError("%s field doesn't exists in %s Model" %
                                       (fields_name[i], Model))

            for row in reader:
                try:
                    obj = Model()
                    for i, field in enumerate(row):
                        setattr(obj, fields_name[i], field)
                    obj.save()
                except Exception as e:
                    raise CommandError(e)



import os
from django.conf import settings
from django.conf.urls import include, url
from django.conf.urls.static import static
from django.contrib import admin
from message_sender import views
from rest_framework.authtoken.views import obtain_auth_token

admin.site.site_header = os.environ.get('MESSAGE_SENDER_TITLE',
                                        'Seed Message Sender Admin')


urlpatterns = [
    url(r'^admin/',  include(admin.site.urls)),
    url(r'^api/auth/',
        include('rest_framework.urls', namespace='rest_framework')),
    url(r'^api/token-auth/', obtain_auth_token),
    url(r'^api/metrics/', views.MetricsView.as_view()),
    url(r'^api/health/', views.HealthcheckView.as_view()),
    url(r'^', include('message_sender.urls')),
    url(r'^docs/', include('rest_framework_docs.urls')),
]


if settings.DEBUG is True:
    urlpatterns += static(
        settings.MEDIA_URL,
        document_root=settings.MEDIA_ROOT,
    )



# -*- coding: utf-8 -*-
# Generated by Django 1.9.12 on 2017-04-19 16:26
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('message_sender', '0012_channel_insert'),
    ]

    operations = [
        migrations.AddField(
            model_name='outbound',
            name='channel',
            field=models.ForeignKey(null=True, on_delete=models.deletion.CASCADE, to='message_sender.Channel'),
        ),
    ]



from django.core.management.base import BaseCommand
from django.core.exceptions import ObjectDoesNotExist
from django.db import connection
import datetime

from message_sender.models import IdentityLookup, Outbound, Inbound

import gc
import pytz


def queryset_iterator_created_at(queryset, chunksize=1000):
    pk = datetime.datetime(2100, 1, 1).replace(tzinfo=pytz.UTC)
    last_pk = queryset.order_by('created_at').values_list(
        'created_at', flat=True).first()
    if last_pk is not None:
        queryset = queryset.order_by('-created_at')
        while pk > last_pk:
            for row in queryset.filter(created_at__lt=pk)[:chunksize]:
                pk = row.created_at
                yield row
            gc.collect()


def queryset_iterator_msisdn(queryset, chunksize=1000):
    pk = ''
    last_pk = queryset.order_by('-msisdn').values_list(
        'msisdn', flat=True).first()
    if last_pk is not None:
        queryset = queryset.order_by('msisdn')
        while pk < last_pk:
            for row in queryset.filter(msisdn__gt=pk)[:chunksize]:
                pk = row.msisdn
                yield row
            gc.collect()


class Command(BaseCommand):
    help = ("This command updates the identity field on the outbound and "
            "inbound table.")

    def add_arguments(self, parser):
        parser.add_argument(
            "--loop", dest="loop", default="ID",
            help=("Loop identities(ID), messages(MSG) or SQL."))

    def handle(self, *args, **options):
        loop = options['loop']

        if loop == "ID":
            self.update_by_id()

        elif loop == "MSG":
            self.update_by_msg()

        elif loop == "SQL":
            self.update_by_sql()

        else:
            self.stdout.write("Invalid Loop(ID, MSG, SQL)")

        self.stdout.write("Updated")

    def update_by_id(self):
        identities = queryset_iterator_msisdn(IdentityLookup.objects.all())

        for identity in identities:

            Outbound.objects.filter(to_addr=identity.msisdn).update(
                to_addr='', to_identity=identity.identity)

            Inbound.objects.filter(from_addr=identity.msisdn).update(
                from_addr='', from_identity=identity.identity)

    def update_by_msg(self):
        outbounds = Outbound.objects.exclude(to_addr='')
        outbounds = queryset_iterator_created_at(outbounds)

        for outbound in outbounds:
            try:
                identity = IdentityLookup.objects.get(
                    msisdn=outbound.to_addr)
                outbound.to_addr = ''
                outbound.to_identity = identity.identity
                outbound.save()
            except ObjectDoesNotExist:
                self.stdout.write("Identity not Found: %s" % outbound.to_addr)

        inbounds = Inbound.objects.exclude(from_addr='')
        inbounds = queryset_iterator_created_at(inbounds)
        for inbound in inbounds:
            try:
                identity = IdentityLookup.objects.get(
                    msisdn=inbound.from_addr)
                inbound.from_addr = ''
                inbound.from_identity = identity.identity
                inbound.save()
            except ObjectDoesNotExist:
                self.stdout.write("Identity not Found: %s" % inbound.to_addr)

    def update_by_sql(self):
        update_out_identity = """
            update message_sender_outbound a
            set to_identity = b.identity
            from
                message_sender_identitylookup b
            where
                a.to_addr = b.msisdn"""

        update_out_addr = """
            update message_sender_outbound
            set to_addr = ''
            where to_identity != '';"""

        update_in_identity = """
            update message_sender_inbound a
            set from_identity = b.identity
            from
                message_sender_identitylookup b
            where
                a.from_addr = b.msisdn"""

        update_in_addr = """
            update message_sender_inbound
            set from_addr = ''
            where from_identity != '';"""

        with connection.cursor() as cursor:
            cursor.execute(update_out_identity)
            cursor.execute(update_out_addr)
            cursor.execute(update_in_identity)
            cursor.execute(update_in_addr)



# -*- coding: utf-8 -*-
# Generated by Django 1.9.12 on 2017-04-19 11:57
from __future__ import unicode_literals
import os

from django.db import migrations
from django.conf import settings


def insert_channels(apps, schema_editor):

    default = True

    Channel = apps.get_model('message_sender.channel')

    if(settings.MESSAGE_BACKEND_TEXT == 'vumi'):
        if not Channel.objects.filter(channel_id='VUMI_TEXT').exists():
            vumi_url = os.environ.get('MESSAGE_SENDER_VUMI_API_URL_TEXT')
            vumi_account_key = os.environ.get(
                'MESSAGE_SENDER_VUMI_ACCOUNT_KEY_TEXT')
            vumi_conv_key = os.environ.get(
                'MESSAGE_SENDER_VUMI_CONVERSATION_KEY_TEXT')
            vumi_acc_token = os.environ.get(
                'MESSAGE_SENDER_VUMI_ACCOUNT_TOKEN_TEXT')

            if (vumi_url and vumi_account_key and vumi_conv_key and
                    vumi_acc_token):
                channel = Channel()
                channel.channel_id = 'VUMI_TEXT'
                channel.default = default
                channel.channel_type = 'vumi'
                channel.concurrency_limit = settings.CONCURRENT_TEXT_LIMIT
                channel.message_delay = settings.TEXT_MESSAGE_DELAY
                channel.message_timeout = settings.TEXT_MESSAGE_TIMEOUT
                channel.configuration = {
                    'VUMI_API_URL': vumi_url,
                    'VUMI_ACCOUNT_KEY': vumi_account_key,
                    'VUMI_CONVERSATION_KEY': vumi_conv_key,
                    'VUMI_ACCOUNT_TOKEN': vumi_acc_token,
                }

                channel.save()

                default = False

    elif(settings.MESSAGE_BACKEND_TEXT == 'junebug'):
        if not Channel.objects.filter(channel_id='JUNEBUG_TEXT').exists():
            june_url = os.environ.get('MESSAGE_SENDER_JUNEBUG_API_URL_TEXT')
            june_auth = os.environ.get('MESSAGE_SENDER_JUNEBUG_API_AUTH_TEXT')
            june_from = os.environ.get('MESSAGE_SENDER_JUNEBUG_API_FROM_TEXT')

            if june_url and june_auth and june_from:
                channel = Channel()
                channel.channel_id = 'JUNEBUG_TEXT'
                channel.default = default
                channel.channel_type = 'junebug'
                channel.concurrency_limit = settings.CONCURRENT_TEXT_LIMIT
                channel.message_delay = settings.TEXT_MESSAGE_DELAY
                channel.message_timeout = settings.TEXT_MESSAGE_TIMEOUT
                channel.configuration = {
                    'JUNEBUG_API_URL': june_url,
                    'JUNEBUG_API_AUTH': june_auth,
                    'JUNEBUG_API_FROM': june_from,
                }
                channel.save()

                default = False

    if (settings.MESSAGE_BACKEND_VOICE == 'vumi'):
        if not Channel.objects.filter(channel_id='VUMI_VOICE').exists():
            vumi_url = os.environ.get('MESSAGE_SENDER_VUMI_API_URL_VOICE')
            vumi_account_key = os.environ.get(
                'MESSAGE_SENDER_VUMI_ACCOUNT_KEY_VOICE')
            vumi_conv_key = os.environ.get(
                'MESSAGE_SENDER_VUMI_CONVERSATION_KEY_VOICE')
            vumi_acc_token = os.environ.get(
                'MESSAGE_SENDER_VUMI_ACCOUNT_TOKEN_VOICE')

            if (vumi_url and vumi_account_key and vumi_conv_key and
                    vumi_acc_token):
                channel = Channel()
                channel.channel_id = 'VUMI_VOICE'
                channel.default = default
                channel.channel_type = 'vumi'
                channel.concurrency_limit = settings.CONCURRENT_VOICE_LIMIT
                channel.message_delay = settings.VOICE_MESSAGE_DELAY
                channel.message_timeout = settings.VOICE_MESSAGE_TIMEOUT
                channel.configuration = {
                    'VUMI_API_URL': vumi_url,
                    'VUMI_ACCOUNT_KEY': vumi_account_key,
                    'VUMI_CONVERSATION_KEY': vumi_conv_key,
                    'VUMI_ACCOUNT_TOKEN': vumi_acc_token,
                }
                channel.save()

                default = False

    elif(settings.MESSAGE_BACKEND_VOICE == 'junebug'):
        if not Channel.objects.filter(channel_id='JUNEBUG_VOICE').exists():
            june_url = os.environ.get('MESSAGE_SENDER_JUNEBUG_API_URL_VOICE')
            june_auth = os.environ.get('MESSAGE_SENDER_JUNEBUG_API_AUTH_VOICE')
            june_from = os.environ.get('MESSAGE_SENDER_JUNEBUG_API_FROM_VOICE')

            if june_url and june_auth and june_from:
                channel = Channel()
                channel.channel_id = 'JUNEBUG_VOICE'
                channel.default = default
                channel.channel_type = 'junebug'
                channel.concurrency_limit = settings.CONCURRENT_VOICE_LIMIT
                channel.message_delay = settings.VOICE_MESSAGE_DELAY
                channel.message_timeout = settings.VOICE_MESSAGE_TIMEOUT
                channel.configuration = {
                    'JUNEBUG_API_URL': june_url,
                    'JUNEBUG_API_AUTH': june_auth,
                    'JUNEBUG_API_FROM': june_from,
                }
                channel.save()


class Migration(migrations.Migration):

    dependencies = [
        ('message_sender', '0011_channel'),
    ]

    operations = [
        migrations.RunPython(insert_channels),
    ]



import json
import os
import re

from copy import deepcopy
from django.conf import settings
from django.core.urlresolvers import reverse

import requests

from go_http.send import HttpApiSender

from .utils import make_absolute_url
from .models import Channel


class FactoryException(Exception):
    pass


class HttpApiSenderException(Exception):
    pass


class GenericHttpApiSender(HttpApiSender):

    def __init__(self, url, auth=None, from_addr=None, session=None,
                 override_payload=None, strip_filepath=False):
        """
        :param url str: The URL for the HTTP API channel
        :param auth tuple: (username, password) or anything
            accepted by the requests library. Defaults to None.
        :param session requests.Session: A requests session. Defaults to None
        :param from_addr str: The from address for all messages. Defaults to
            None
        :param override_payload dict: This is the format of the payload that
            needs to be sent to the URL. It willl be populated from the
            original payload. Defaults to None
        :param strip_filepath boolean: This should be true if we only need to
            send the filename to the api.
        """
        self.api_url = url
        self.auth = tuple(auth) if isinstance(auth, list) else auth
        self.from_addr = from_addr
        if session is None:
            session = requests.Session()
        self.session = session
        self.override_payload = override_payload
        self.strip_filepath = strip_filepath

    def _get_filename(self, path):
        """
        This function gets the base filename from the path, if a language code
        is present the filename will start from there.
        """
        match = re.search('[a-z]{2,3}_[A-Z]{2}', path)

        if match:
            start = match.start(0)
            filename = path[start:]
        else:
            filename = os.path.basename(path)

        return filename

    def _raw_send(self, py_data):
        headers = {'content-type': 'application/json; charset=utf-8'}

        channel_data = py_data.get('helper_metadata', {})
        channel_data['session_event'] = py_data.get('session_event')

        url = channel_data.get('voice', {}).get('speech_url')
        if self.strip_filepath and url:
            if not isinstance(url, (list, tuple)):
                channel_data['voice']['speech_url'] = self._get_filename(url)
            else:
                channel_data['voice']['speech_url'] = []
                for item in url:
                    channel_data['voice']['speech_url'].append(
                        self._get_filename(item))

        data = {
            'to': py_data['to_addr'],
            'from': self.from_addr,
            'content': py_data['content'],
            'channel_data': channel_data
        }

        data = self._override_payload(data)

        data = json.dumps(data)
        r = self.session.post(self.api_url, auth=self.auth,
                              data=data, headers=headers,
                              timeout=settings.DEFAULT_REQUEST_TIMEOUT)
        r.raise_for_status()
        res = r.json()
        return res.get('result', {})

    def _override_payload(self, payload):
        """
        This function transforms the payload into a new format using the
        self.override_payload property.
        """
        if self.override_payload:
            old_payload = payload

            def get_value(data, key):
                try:
                    parent_key, nested_key = key.split('.', 1)
                    return get_value(data.get(parent_key, {}), nested_key)
                except ValueError:
                    return data.get(key, key)

            def set_values(data):
                for key, value in data.items():
                    if isinstance(value, dict):
                        set_values(value)
                    else:
                        data[key] = get_value(old_payload, value)

            payload = deepcopy(self.override_payload)
            set_values(payload)

        return payload

    def fire_metric(self, metric, value, agg="last"):
        raise HttpApiSenderException(
            'Metrics sending not supported')


class JunebugApiSender(GenericHttpApiSender):

    def _raw_send(self, py_data):
        headers = {'content-type': 'application/json; charset=utf-8'}

        channel_data = py_data.get('helper_metadata', {})
        channel_data['session_event'] = py_data.get('session_event')

        data = {
            'to': py_data['to_addr'],
            'from': self.from_addr,
            'content': py_data['content'],
            'channel_data': channel_data,
            'event_url': make_absolute_url(reverse('junebug-events')),
        }

        data = json.dumps(data)
        r = self.session.post(self.api_url, auth=self.auth,
                              data=data, headers=headers,
                              timeout=settings.DEFAULT_REQUEST_TIMEOUT)
        r.raise_for_status()
        res = r.json()
        return res.get('result', {})


class MessageClientFactory(object):

    @classmethod
    def create(cls, channel=None):
        try:
            if not channel:
                channel = Channel.objects.get(default=True)
        except Channel.DoesNotExist:
            raise FactoryException(
                'Unknown backend type: %r' % (channel,))

        backend_type = channel.channel_type
        handler = getattr(cls,
                          'create_%s_client' % (backend_type,), None)
        if not handler:
            raise FactoryException(
                'Unknown backend type: %r' % (backend_type,))

        return handler(channel)

    @classmethod
    def create_junebug_client(cls, channel):
        return JunebugApiSender(
            channel.configuration.get("JUNEBUG_API_URL"),
            channel.configuration.get("JUNEBUG_API_AUTH"),
            channel.configuration.get("JUNEBUG_API_FROM")
        )

    @classmethod
    def create_vumi_client(cls, channel):
        return HttpApiSender(
            channel.configuration.get("VUMI_ACCOUNT_KEY"),
            channel.configuration.get("VUMI_CONVERSATION_KEY"),
            channel.configuration.get("VUMI_ACCOUNT_TOKEN"),
            api_url=channel.configuration.get("VUMI_API_URL")
        )

    @classmethod
    def create_http_api_client(cls, channel):
        return GenericHttpApiSender(
            channel.configuration.get("HTTP_API_URL"),
            channel.configuration.get("HTTP_API_AUTH"),
            channel.configuration.get("HTTP_API_FROM"),
            override_payload=channel.configuration.get("OVERRIDE_PAYLOAD"),
            strip_filepath=channel.configuration.get("STRIP_FILEPATH"),
        )



# -*- coding: utf-8 -*-
# Generated by Django 1.9.12 on 2017-04-19 11:57
from __future__ import unicode_literals

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('message_sender', '0010_merge'),
    ]

    operations = [
        migrations.CreateModel(
            name='Channel',
            fields=[
                ('channel_id', models.CharField(editable=True, max_length=64, primary_key=True, serialize=False)),
                ('channel_type', models.CharField(choices=[(b'junebug', b'Junebug'), (b'vumi', b'Vumi')], default=b'junebug', max_length=20)),
                ('concurrency_limit', models.IntegerField(default=0)),
                ('message_delay', models.IntegerField(default=0)),
                ('message_timeout', models.IntegerField(default=0)),
                ('default', models.BooleanField(default=False)),
                ('configuration', django.contrib.postgres.fields.jsonb.JSONField()),
            ],
        ),
    ]



from .models import Inbound, Outbound, OutboundSendFailure, Channel
from rest_hooks.models import Hook
from rest_framework import serializers


class OneFieldRequiredValidator:
    def __init__(self, fields):
        self.fields = fields

    def set_context(self, serializer):
        self.is_create = getattr(serializer, 'instance', None) is None

    def __call__(self, data):
        if self.is_create:

            for field in self.fields:
                if data.get(field):
                    return

            raise serializers.ValidationError(
                "One of these fields must be populated: %s" %
                (', '.join(self.fields)))


class OutboundSerializer(serializers.HyperlinkedModelSerializer):

    channel = serializers.PrimaryKeyRelatedField(
        queryset=Channel.objects.all(), required=False)

    class Meta:
        model = Outbound
        fields = (
            'url', 'id', 'version', 'to_addr', 'vumi_message_id', 'content',
            'delivered', 'attempts', 'metadata', 'created_at', 'updated_at',
            'channel', 'to_identity', 'resend')
        validators = [OneFieldRequiredValidator(['to_addr', 'to_identity'])]


class OutboundArchiveSerializer(serializers.ModelSerializer):
    class Meta:
        model = Outbound
        fields = (
            'id', 'to_addr', 'to_identity', 'version', 'content',
            'vumi_message_id', 'delivered', 'resend', 'call_answered',
            'attempts', 'metadata', 'channel', 'updated_at', 'created_at',
            'last_sent_time', 'created_by', 'updated_by',
        )


class InboundSerializer(serializers.HyperlinkedModelSerializer):

    class Meta:
        model = Inbound
        fields = (
            'url', 'id', 'message_id', 'in_reply_to', 'to_addr',
            'from_addr', 'content', 'transport_name', 'transport_type',
            'helper_metadata', 'created_at', 'updated_at', 'from_identity')
        validators = \
            [OneFieldRequiredValidator(['from_addr', 'from_identity'])]

    def to_internal_value(self, data):
        """
        Adds extra data to the helper_metadata field.
        """
        if "session_event" in data:
            data['helper_metadata']['session_event'] = data['session_event']

        return super(InboundSerializer, self).to_internal_value(data)


class JunebugInboundSerializer(serializers.HyperlinkedModelSerializer):
    """
    Maps fields from Junebug onto fields expected by the Inbound model.
    """
    reply_to = serializers.CharField(source='in_reply_to')
    to = serializers.CharField(source='to_addr')
    channel_id = serializers.CharField(source='transport_name')
    channel_data = serializers.JSONField(source='helper_metadata')

    class Meta:
        model = Inbound
        fields = (
            'url', 'id', 'message_id', 'reply_to', 'to', 'from_addr',
            'content', 'channel_id', 'channel_data', 'created_at',
            'updated_at', 'from_identity')
        validators = \
            [OneFieldRequiredValidator(['from', 'from_identity'])]


class HookSerializer(serializers.ModelSerializer):

    class Meta:
        model = Hook
        read_only_fields = ('user',)


class CreateUserSerializer(serializers.Serializer):
    email = serializers.EmailField()


class OutboundSendFailureSerializer(serializers.HyperlinkedModelSerializer):

    class Meta:
        model = OutboundSendFailure
        fields = ('url', 'id', 'outbound', 'task_id', 'initiated_at', 'reason')


class AggregateOutboundSerializer(serializers.Serializer):
    start = serializers.DateField(required=False)
    end = serializers.DateField(required=False)


class ArchivedOutboundSerializer(serializers.Serializer):
    start = serializers.DateField()
    end = serializers.DateField()



import gzip
import json
import os
import pytz
import random
import requests
import time

from celery.exceptions import MaxRetriesExceededError
from celery.task import Task
from celery.utils.log import get_task_logger

from datetime import datetime, timedelta
from django.conf import settings
from django.core.cache import cache
from django.core.exceptions import ObjectDoesNotExist
from django.core.files import File
from django.db.models import Count, Sum

from seed_services_client.metrics import MetricsApiClient
from requests import exceptions as requests_exceptions
from rest_framework.renderers import JSONRenderer

from .factory import MessageClientFactory


from .models import (
    Outbound, OutboundSendFailure, Channel, AggregateOutbounds,
    ArchivedOutbounds,
)
from .serializers import OutboundArchiveSerializer
from seed_message_sender.utils import (
    load_callable, get_identity_address, get_identity_by_address,
    create_identity)
from message_sender.utils import daterange
from seed_papertrail.decorators import papertrail

logger = get_task_logger(__name__)

voice_to_addr_formatter = load_callable(settings.VOICE_TO_ADDR_FORMATTER)
text_to_addr_formatter = load_callable(settings.TEXT_TO_ADDR_FORMATTER)


def calculate_retry_delay(attempt, max_delay=300):
    """Calculates an exponential backoff for retry attempts with a small
    amount of jitter."""
    delay = int(random.uniform(2, 4) ** attempt)
    if delay > max_delay:
        # After reaching the max delay, stop using expontential growth
        # and keep the delay nearby the max.
        delay = int(random.uniform(max_delay - 20, max_delay + 20))
    return delay


class DeliverHook(Task):
    def run(self, target, payload, instance_id=None, hook_id=None, **kwargs):
        """
        target:     the url to receive the payload.
        payload:    a python primitive data structure
        instance_id:   a possibly None "trigger" instance ID
        hook_id:       the ID of defining Hook object
        """
        requests.post(
            url=target,
            data=json.dumps(payload),
            headers={
                'Content-Type': 'application/json',
                'Authorization': 'Token %s' % settings.HOOK_AUTH_TOKEN
            }
        )


def deliver_hook_wrapper(target, payload, instance, hook):
    if instance is not None:
        instance_id = instance.id
    else:
        instance_id = None
    kwargs = dict(target=target, payload=payload,
                  instance_id=instance_id, hook_id=hook.id)
    DeliverHook.apply_async(kwargs=kwargs)


def get_metric_client(session=None):
    return MetricsApiClient(
        url=settings.METRICS_URL,
        auth=settings.METRICS_AUTH,
        session=session)


class FireMetric(Task):

    """ Fires a metric using the MetricsApiClient
    """
    name = "message_sender.tasks.fire_metric"

    @papertrail.debug(name, sample=0.1)
    def run(self, metric_name, metric_value, session=None, **kwargs):
        metric_value = float(metric_value)
        metric = {
            metric_name: metric_value
        }
        metric_client = get_metric_client(session=session)
        metric_client.fire_metrics(**metric)
        return "Fired metric <%s> with value <%s>" % (
            metric_name, metric_value)


fire_metric = FireMetric()


class ConcurrencyLimiter(object):
    BUCKET_SIZE = 60

    @classmethod
    def get_key(cls, channel_id, bucket):
        return "%s_messages_at_%s" % (channel_id, bucket)

    @classmethod
    def get_current_message_count(cls, channel):
        # Sum the values in all the buckets to get the total
        total = 0
        number_of_buckets = channel.message_timeout // cls.BUCKET_SIZE + 1
        bucket = int(time.time() // cls.BUCKET_SIZE)
        for i in range(bucket, bucket - number_of_buckets, -1):
            value = cache.get(cls.get_key(channel.channel_id, i))
            if value:
                total += int(value)
        return total

    @classmethod
    def incr_message_count(cls, channel_id, timeout):
        bucket = int(time.time() // cls.BUCKET_SIZE)
        key = cls.get_key(channel_id, bucket)

        # Add the bucket size to the expiry time so messages that start at
        # the end of the bucket still complete
        if not cache.add(key, 1, timeout + cls.BUCKET_SIZE):
            cache.incr(key)

    @classmethod
    def decr_message_count(cls, channel, msg_time):

        if channel.concurrency_limit == 0:
            return
        timeout = channel.message_timeout

        if not msg_time:
            return

        # Convert from datetime to seconds since epoch
        msg_time = msg_time.replace(tzinfo=None) - msg_time.utcoffset()
        msg_time = (msg_time - datetime(1970, 1, 1)).total_seconds()

        time_since = time.time() - msg_time
        if time_since > timeout:
            return
        bucket = int(msg_time // cls.BUCKET_SIZE)

        key = cls.get_key(channel.channel_id, bucket)
        # Set the expiry time to the timeout minus the time passed since
        # the message was sent.
        if int(cache.get_or_set(key, lambda: 0, timeout - time_since)) > 0:
            cache.decr(key)

    @classmethod
    def manage_limit(cls, task, channel):
        limit = channel.concurrency_limit
        timeout = channel.message_timeout
        delay = channel.message_delay

        if limit > 0:
            if cls.get_current_message_count(channel) >= limit:
                task.retry(countdown=delay)
            cls.incr_message_count(channel.channel_id, timeout)


class SendMessage(Task):

    """
    Task to load and contruct message and send them off
    """
    name = "message_sender.tasks.send_message"
    default_retry_delay = 5
    max_retries = None
    max_error_retries = 5

    class FailedEventRequest(Exception):

        """
        The attempted task failed because of a non-200 HTTP return
        code.
        """

    def get_client(self, channel=None):
        return MessageClientFactory.create(channel)

    @papertrail.debug(name, sample=0.1)
    def run(self, message_id, **kwargs):
        """
        Load and contruct message and send them off
        """
        l = self.get_logger(**kwargs)

        error_retry_count = kwargs.get('error_retry_count', 0)
        if error_retry_count >= self.max_error_retries:
            raise MaxRetriesExceededError(
                "Can't retry {0}[{1}] args:{2} kwargs:{3}".format(
                    self.name, self.request.id, self.request.args, kwargs))

        l.info("Loading Outbound Message <%s>" % message_id)
        try:
            message = Outbound.objects.select_related('channel').get(
                id=message_id)
        except ObjectDoesNotExist:
            logger.error('Missing Outbound message', exc_info=True)
            return

        if message.attempts < settings.MESSAGE_SENDER_MAX_RETRIES:
            if error_retry_count > 0:
                retry_delay = calculate_retry_delay(error_retry_count)
            else:
                retry_delay = self.default_retry_delay
            l.info("Attempts: %s" % message.attempts)
            # send or resend
            try:
                if not message.channel:
                    channel = Channel.objects.get(default=True)
                else:
                    channel = message.channel

                sender = self.get_client(channel)
                ConcurrencyLimiter.manage_limit(self, channel)

                if not message.to_addr and message.to_identity:
                    message.to_addr = get_identity_address(
                        message.to_identity, use_communicate_through=True)

                if message.to_addr and not message.to_identity:
                    result = get_identity_by_address(message.to_addr)

                    if result:
                        message.to_identity = result[0]['id']
                    else:
                        identity = {
                            'details': {
                                'default_addr_type': 'msisdn',
                                'addresses': {
                                    'msisdn': {
                                        message.to_addr: {'default': True}
                                    }
                                }
                            }
                        }
                        identity = create_identity(identity)
                        message.to_identity = identity['id']

                if "voice_speech_url" in message.metadata:
                    # OBD number of tries metric
                    fire_metric.apply_async(kwargs={
                        "metric_name": 'vumimessage.obd.tries.sum',
                        "metric_value": 1.0
                    })

                    # Voice message
                    speech_url = message.metadata["voice_speech_url"]
                    vumiresponse = sender.send_voice(
                        voice_to_addr_formatter(message.to_addr),
                        message.content,
                        speech_url=speech_url,
                        session_event="new")
                    l.info("Sent voice message to <%s>" % message.to_addr)
                else:
                    # Plain content
                    vumiresponse = sender.send_text(
                        text_to_addr_formatter(message.to_addr),
                        message.content,
                        session_event="new")
                    l.info("Sent text message to <%s>" % (
                        message.to_addr,))

                message.last_sent_time = datetime.now()
                message.attempts += 1
                message.vumi_message_id = vumiresponse["message_id"]
                message.save()
                fire_metric.apply_async(kwargs={
                    "metric_name": 'vumimessage.tries.sum',
                    "metric_value": 1.0
                })
            except requests_exceptions.ConnectionError as exc:
                l.info('Connection Error sending message')
                fire_metric.delay(
                    'sender.send_message.connection_error.sum', 1)
                kwargs['error_retry_count'] = error_retry_count + 1
                self.retry(exc=exc, countdown=retry_delay, args=(message_id,),
                           kwargs=kwargs)
            except requests_exceptions.Timeout as exc:
                l.info('Sending message failed due to timeout')
                fire_metric.delay('sender.send_message.timeout.sum', 1)
                kwargs['error_retry_count'] = error_retry_count + 1
                self.retry(exc=exc, countdown=retry_delay, args=(message_id,),
                           kwargs=kwargs)
            except requests_exceptions.HTTPError as exc:
                # retry message sending if in 500 range (3 default
                # retries)
                l.info('Sending message failed due to status: %s' %
                       exc.response.status_code)
                metric_name = ('sender.send_message.http_error.%s.sum' %
                               exc.response.status_code)
                fire_metric.delay(metric_name, 1)
                kwargs['error_retry_count'] = error_retry_count + 1
                self.retry(exc=exc, countdown=retry_delay, args=(message_id,),
                           kwargs=kwargs)

            # If we've gotten this far the message send was successful.
            fire_metric.apply_async(kwargs={
                "metric_name": 'message.sent.sum',
                "metric_value": 1.0
            })
            return vumiresponse

        else:
            # This is for retries based on async nacks from the transport.
            l.info("Message <%s> at max retries." % str(message_id))
            message.to_addr = ''
            message.save(update_fields=['to_addr'])
            fire_metric.apply_async(kwargs={
                "metric_name": 'vumimessage.maxretries.sum',
                "metric_value": 1.0
            })
            # Count failures on exhausted tries.
            fire_metric.apply_async(kwargs={
                "metric_name": 'message.failures.sum',
                "metric_value": 1.0
            })

    def on_failure(self, exc, task_id, args, kwargs, einfo):
        error_retry_count = kwargs.get('error_retry_count', 0)
        if error_retry_count == self.max_error_retries:
            if 'message_id' in kwargs:
                message_id = kwargs['message_id']
            else:
                message_id = args[0]
            OutboundSendFailure.objects.create(
                outbound_id=message_id,
                initiated_at=self.request.eta,
                reason=einfo.exception.message,
                task_id=task_id
            )
            # Count permanent failures.
            fire_metric.apply_async(kwargs={
                "metric_name": 'message.failures.sum',
                "metric_value": 1.0
            })
        super(SendMessage, self).on_failure(exc, task_id, args,
                                            kwargs, einfo)


send_message = SendMessage()


class RequeueFailedTasks(Task):

    """
    Task to requeue failed Outbounds.
    """
    name = "message_sender.tasks.requeue_failed_tasks"

    def run(self, **kwargs):
        l = self.get_logger(**kwargs)
        failures = OutboundSendFailure.objects
        l.info("Attempting to requeue <%s> failed Outbound sends" %
               failures.all().count())
        for failure in failures.iterator():
            outbound_id = str(failure.outbound_id)
            # Cleanup the failure before requeueing it.
            failure.delete()
            send_message.delay(outbound_id)


requeue_failed_tasks = RequeueFailedTasks()


class AggregateOutboundMessages(Task):
    """
    Task to aggregate the outbound messages and store the results in the
    aggregate table
    """
    name = "message_sender.tasks.aggregate_outbounds"

    def run(self, start_date, end_date):
        start_date = datetime.strptime(
            start_date, '%Y-%m-%d').replace(tzinfo=pytz.UTC)
        end_date = datetime.strptime(
            end_date, '%Y-%m-%d').replace(tzinfo=pytz.UTC)

        # Delete any existing aggregates for these dates. This is necessary
        # to avoid having leftovers from changed objects. eg. There were
        # undelivered messages, but now they're all delivered, so we don't want
        # the undelivered aggregate to still be there, but an update won't set
        # the undelivered aggregate to 0.
        AggregateOutbounds.objects.filter(
            date__gte=start_date.date(), date__lte=end_date.date()).delete()

        for d in daterange(start_date, end_date):
            query = Outbound.objects.filter(
                created_at__gte=d,
                created_at__lt=(d + timedelta(1))
            )
            query = query.values('delivered', 'channel')
            query = query.annotate(attempts=Sum('attempts'), total=Count('*'))
            for aggregate in query.iterator():
                AggregateOutbounds.objects.create(
                    date=d,
                    delivered=aggregate['delivered'],
                    channel_id=aggregate['channel'],
                    attempts=aggregate['attempts'],
                    total=aggregate['total'],
                )

aggregate_outbounds = AggregateOutboundMessages()


class ArchiveOutboundMessages(Task):
    """
    Task to archive the outbound messages and store the messages in the
    storage backend
    """
    name = "message_sender.tasks.archive_outbounds"

    def filename(self, date):
        """
        Returns the filename for the provided date
        """
        return 'outbounds-{}.gz'.format(date.strftime('%Y-%m-%d'))

    def dump_data(self, filename, queryset):
        """
        Serializes the queryset into a newline separated JSON format, and
        places it into a gzipped file
        """
        with gzip.open(filename, 'wb') as f:
            for outbound in queryset.iterator():
                data = OutboundArchiveSerializer(outbound).data
                data = JSONRenderer().render(data)
                f.write(data)
                f.write('\n'.encode('utf-8'))

    def create_archived_outbound(self, date, filename):
        """
        Creates the required ArchivedOutbound entry with the file specified
        at `filename`
        """
        with open(filename, 'rb') as f:
            f = File(f)
            ArchivedOutbounds.objects.create(date=date, archive=f)

    def run(self, start_date, end_date):
        start_date = datetime.strptime(
            start_date, '%Y-%m-%d').replace(tzinfo=pytz.UTC)
        end_date = datetime.strptime(
            end_date, '%Y-%m-%d').replace(tzinfo=pytz.UTC)

        for d in daterange(start_date, end_date):
            if ArchivedOutbounds.objects.filter(date=d.date()).exists():
                continue

            query = Outbound.objects.filter(
                created_at__gte=d,
                created_at__lt=(d + timedelta(1))
            )

            if not query.exists():
                continue

            filename = self.filename(d)
            self.dump_data(filename, query)
            self.create_archived_outbound(d, filename)

            os.remove(filename)
            query.delete()

archive_outbound = ArchiveOutboundMessages()



# -*- coding: utf-8 -*-
# Generated by Django 1.9.12 on 2018-02-20 13:29
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('message_sender', '0016_outbound_resend'),
    ]

    operations = [
        migrations.AlterField(
            model_name='channel',
            name='channel_type',
            field=models.CharField(choices=[(b'junebug', b'Junebug'), (b'vumi', b'Vumi'), (b'http_api', b'HTTP API')], default=b'junebug', max_length=20),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.12 on 2018-03-16 13:05
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('message_sender', '0017_auto_20180220_1329'),
    ]

    operations = [
        migrations.CreateModel(
            name='AggregateOutbounds',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date', models.DateField(help_text=b'The date that the aggregate is for')),
                ('delivered', models.BooleanField(help_text=b'Whether this is for delivery passed or failed messages')),
                ('attempts', models.IntegerField(help_text=b'The total number of attempts')),
                ('total', models.IntegerField(help_text=b'The total number of messages')),
                ('channel', models.ForeignKey(help_text=b'Which channel this is for', null=True, on_delete=django.db.models.deletion.SET_NULL, to='message_sender.Channel')),
            ],
        ),
        migrations.AlterUniqueTogether(
            name='aggregateoutbounds',
            unique_together=set([('date', 'delivered', 'channel')]),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.12 on 2018-01-31 10:22
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('message_sender', '0015_identitylookup'),
    ]

    operations = [
        migrations.AddField(
            model_name='outbound',
            name='resend',
            field=models.NullBooleanField(default=None, help_text=b'True if this is a resend requested by the user.'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.12 on 2017-05-08 14:00
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('message_sender', '0014_auto_20170503_1248'),
    ]

    operations = [
        migrations.CreateModel(
            name='IdentityLookup',
            fields=[
                ('msisdn', models.CharField(max_length=255, primary_key=True, serialize=False)),
                ('identity', models.CharField(blank=False, max_length=36, null=False)),
            ],
        ),
    ]
